#include "VulkanHeaders.h"
#include "MixedSampleCountDynamicRendering.h"

#include "renderdoc_app.h"

#include <vector>

#define NOMINMAX
#include <Windows.h>
#undef MemoryBarrier

inline RENDERDOC_API_1_5_0* rdoc_api = nullptr;

void main_mixed_sample_count_dynamic_rendering(const std::int32_t argc, const char** argv) {

	if (HMODULE mod = GetModuleHandleA("renderdoc.dll")) {
		pRENDERDOC_GetAPI RENDERDOC_GetAPI = (pRENDERDOC_GetAPI) GetProcAddress(mod, "RENDERDOC_GetAPI");
		RENDERDOC_GetAPI(eRENDERDOC_API_Version_1_5_0, (void**) &rdoc_api);
	}

	///////////////////////////////////
	//        Dynamic Dispatch       //
	///////////////////////////////////
	vk::DynamicLoader loader{};

	{

		PFN_vkGetInstanceProcAddr vkGetInstanceProcAddr = loader.getProcAddress<PFN_vkGetInstanceProcAddr>("vkGetInstanceProcAddr");

		VULKAN_HPP_DEFAULT_DISPATCHER.init(vkGetInstanceProcAddr);

	}

	///////////////////////////////////
	//           Instance            //
	///////////////////////////////////
	vk::Instance instance{};
	vk::DebugUtilsMessengerEXT debug_messenger{};

	{

		constexpr const char* extensions[] = {
			VK_EXT_DEBUG_UTILS_EXTENSION_NAME
		};

		constexpr const char* layers[] = {
			"VK_LAYER_KHRONOS_validation"
		};

		constexpr std::array enables = {
			vk::ValidationFeatureEnableEXT::eGpuAssisted,
			vk::ValidationFeatureEnableEXT::eSynchronizationValidation,
		};

		const vk::ApplicationInfo app_info = {
			.applicationVersion = VK_MAKE_API_VERSION(0, 0, 1, 0),
			.pEngineName = "NvidiaBugs",
			.engineVersion = VK_MAKE_API_VERSION(0, 0, 1, 0),
			.apiVersion = VK_API_VERSION_1_3
		};

		const vk::ValidationFeaturesEXT features = {
			.enabledValidationFeatureCount = static_cast<std::uint32_t>(enables.size()),
			.pEnabledValidationFeatures = enables.data(),
		};

		instance = vk::createInstance(vk::InstanceCreateInfo{
#if defined(DEBUG) || defined(_DEBUG)
			.pNext = &features,
#endif
			.pApplicationInfo = &app_info,
#if defined(DEBUG) || defined(_DEBUG)
			.enabledLayerCount = static_cast<std::uint32_t>(std::size(layers)),
			.ppEnabledLayerNames = layers,
			.enabledExtensionCount = static_cast<std::uint32_t>(std::size(extensions)),
			.ppEnabledExtensionNames = extensions,
#endif
		});

		VULKAN_HPP_DEFAULT_DISPATCHER.init(instance);
#if defined(DEBUG) || defined(_DEBUG)
		const vk::DebugUtilsMessengerCreateInfoEXT debug_info = {
			.messageSeverity = vk::DebugUtilsMessageSeverityFlagBitsEXT::eVerbose |
							   vk::DebugUtilsMessageSeverityFlagBitsEXT::eError |
							   vk::DebugUtilsMessageSeverityFlagBitsEXT::eWarning,
			.messageType = vk::DebugUtilsMessageTypeFlagBitsEXT::eGeneral |
						   vk::DebugUtilsMessageTypeFlagBitsEXT::ePerformance |
						   vk::DebugUtilsMessageTypeFlagBitsEXT::eValidation,
			.pfnUserCallback = debug_callback
		};

		debug_messenger = instance.createDebugUtilsMessengerEXT(debug_info);
#endif
	}

	///////////////////////////////////
	//        Physcical Device       //
	///////////////////////////////////
	vk::PhysicalDevice physical_device = instance.enumeratePhysicalDevices()[0];

	///////////////////////////////////
	//            Device             //
	///////////////////////////////////
	vk::Device device{};
	vk::Queue graphics_queue{};
	std::uint32_t graphics_queue_index{};

	{

		const std::vector<vk::QueueFamilyProperties> queue_families = physical_device.getQueueFamilyProperties();

		const auto findBestQueue = [&](const vk::QueueFlagBits bits) -> std::int32_t {
			const auto bestFinder = [&](const vk::QueueFamilyProperties& lhs,
										const vk::QueueFamilyProperties& rhs) noexcept -> bool {
				if (lhs.queueFlags & bits &&
					!(rhs.queueFlags & bits)) {
					return false;
				} else if (rhs.queueFlags & bits &&
						   !(lhs.queueFlags & bits)) {
					return true;
				}

				return std::popcount(static_cast<std::uint32_t>(lhs.queueFlags)) <
					   std::popcount(static_cast<std::uint32_t>(rhs.queueFlags));
			};

			const auto leastFinder = [&](const vk::QueueFamilyProperties& lhs,
										 const vk::QueueFamilyProperties& rhs) noexcept -> bool {
				if (lhs.queueFlags & bits &&
					!(rhs.queueFlags & bits)) {
					return false;
				} else if (rhs.queueFlags & bits &&
						   !(lhs.queueFlags & bits)) {
					return true;
				}

				return !(std::popcount(static_cast<std::uint32_t>(lhs.queueFlags)) <
						 std::popcount(static_cast<std::uint32_t>(rhs.queueFlags)));
			};

			if (bits == vk::QueueFlagBits::eGraphics) {
				const auto iterator = std::max_element(queue_families.begin(), queue_families.end(), bestFinder);
				return static_cast<std::int32_t>(std::distance(queue_families.begin(), iterator));
			} else {
				const auto iterator = std::max_element(queue_families.begin(), queue_families.end(), leastFinder);
				return static_cast<std::int32_t>(std::distance(queue_families.begin(), iterator));
			}
		};

		graphics_queue_index = findBestQueue(vk::QueueFlagBits::eGraphics);

		constexpr const char* extensions[] = {
			VK_NV_FRAMEBUFFER_MIXED_SAMPLES_EXTENSION_NAME
		};

		vk::StructureChain<vk::PhysicalDeviceFeatures2, vk::PhysicalDeviceVulkan12Features, vk::PhysicalDeviceVulkan13Features> features;

		std::get<0>(features).features.independentBlend = VK_TRUE;
		std::get<1>(features).drawIndirectCount = VK_TRUE;
		std::get<2>(features).dynamicRendering = VK_TRUE;
		std::get<2>(features).synchronization2 = VK_TRUE;

		constexpr float priority = 1.0f;

		const vk::DeviceQueueCreateInfo queue_create_info = {
			.queueFamilyIndex = graphics_queue_index,
			.queueCount = 1,
			.pQueuePriorities = &priority,
		};

		const vk::DeviceCreateInfo createInfo = {
			.pNext = &std::get<0>(features),
			.queueCreateInfoCount = 1,
			.pQueueCreateInfos = &queue_create_info,
			.enabledLayerCount = 0,
			.enabledExtensionCount = static_cast<std::uint32_t>(std::size(extensions)),
			.ppEnabledExtensionNames = extensions
		};

		device = physical_device.createDevice(createInfo);
		graphics_queue = device.getQueue(graphics_queue_index, 0);

		VULKAN_HPP_DEFAULT_DISPATCHER.init(instance, loader.getProcAddress<PFN_vkGetInstanceProcAddr>("vkGetInstanceProcAddr"),
										   device, loader.getProcAddress<PFN_vkGetDeviceProcAddr>("vkGetDeviceProcAddr"));

	}

	///////////////////////////////////
	//          Command Pool         //
	///////////////////////////////////
	vk::CommandPool command_pool = device.createCommandPool(vk::CommandPoolCreateInfo{
		.flags = vk::CommandPoolCreateFlagBits::eResetCommandBuffer,
		.queueFamilyIndex = graphics_queue_index
	});

	///////////////////////////////////
	//         Command Buffer        //
	///////////////////////////////////
	vk::CommandBuffer command_buffer = device.allocateCommandBuffers(vk::CommandBufferAllocateInfo{
		.commandPool = command_pool,
		.level = vk::CommandBufferLevel::ePrimary,
		.commandBufferCount = 1u
	})[0];

	///////////////////////////////////
	//         Attachments	         //
	///////////////////////////////////
	std::uint32_t attachment_width = 1080u;
	std::uint32_t attachment_height = 603u;

	vk::Image attachments[5]{};
	vk::DeviceMemory memories[5]{};
	vk::ImageView attachment_views[5]{};

	vk::ColorComponentFlags attachment_color_components[4] = {
		vk::ColorComponentFlagBits::eR | vk::ColorComponentFlagBits::eG | vk::ColorComponentFlagBits::eB,
		vk::ColorComponentFlagBits::eR | vk::ColorComponentFlagBits::eG | vk::ColorComponentFlagBits::eB | vk::ColorComponentFlagBits::eA,
		vk::ColorComponentFlagBits::eR | vk::ColorComponentFlagBits::eG,
		vk::ColorComponentFlagBits::eR | vk::ColorComponentFlagBits::eG | vk::ColorComponentFlagBits::eB | vk::ColorComponentFlagBits::eA
	};

	vk::Format attachment_formats[5] = {
		vk::Format::eD32Sfloat,
		vk::Format::eB10G11R11UfloatPack32,
		vk::Format::eA2B10G10R10UnormPack32,
		vk::Format::eR16G16Sfloat,
		vk::Format::eR8G8B8A8Unorm,
	};

	vk::SampleCountFlagBits attachment_samples[5] = {
		vk::SampleCountFlagBits::e8,
		vk::SampleCountFlagBits::e8,
		vk::SampleCountFlagBits::e1,
		vk::SampleCountFlagBits::e1,
		vk::SampleCountFlagBits::e1
	};

	{

		{

			const vk::ImageCreateInfo image_create_info = {
				.imageType = vk::ImageType::e2D,
				.format = attachment_formats[0],
				.extent = { attachment_width, attachment_height, 1u },
				.mipLevels = 1u,
				.arrayLayers = 1u,
				.samples = attachment_samples[0],
				.usage = vk::ImageUsageFlagBits::eDepthStencilAttachment
			};

			const vk::MemoryRequirements2 result = device.getImageMemoryRequirements(vk::DeviceImageMemoryRequirements{
				.pCreateInfo = &image_create_info,
				.planeAspect = vk::ImageAspectFlagBits::eDepth
			});

			const auto index = find_memory_type(physical_device, result.memoryRequirements.memoryTypeBits, vk::MemoryPropertyFlagBits::eDeviceLocal);

			memories[0] = device.allocateMemory(vk::MemoryAllocateInfo{
				.allocationSize = std::max<vk::DeviceSize>(attachment_width * attachment_height * 4, result.memoryRequirements.size),
				.memoryTypeIndex = index
			});

			attachments[0] = device.createImage(image_create_info);

			device.bindImageMemory(attachments[0], memories[0], 0);

			attachment_views[0] = device.createImageView(vk::ImageViewCreateInfo{
				.image = attachments[0],
				.viewType = vk::ImageViewType::e2D,
				.format = attachment_formats[0],
				.subresourceRange = vk::ImageSubresourceRange{
					.aspectMask = vk::ImageAspectFlagBits::eDepth,
					.baseMipLevel = 0u,
					.levelCount = 1u,
					.baseArrayLayer = 0u,
					.layerCount = 1u
				}
			});
		}

		{

			for (std::uint32_t i = 1u; i < 5u; i++) {

				const vk::ImageCreateInfo image_create_info = {
					.imageType = vk::ImageType::e2D,
					.format = attachment_formats[i],
					.extent = { attachment_width, attachment_height, 1u },
					.mipLevels = 1u,
					.arrayLayers = 1u,
					.samples = attachment_samples[i],
					.usage = vk::ImageUsageFlagBits::eColorAttachment
				};

				const vk::MemoryRequirements2 result = device.getImageMemoryRequirements(vk::DeviceImageMemoryRequirements{
					.pCreateInfo = &image_create_info,
					.planeAspect = vk::ImageAspectFlagBits::eColor
				});

				const auto index = find_memory_type(physical_device, result.memoryRequirements.memoryTypeBits, vk::MemoryPropertyFlagBits::eDeviceLocal);

				memories[i] = device.allocateMemory(vk::MemoryAllocateInfo{
					.allocationSize = std::max<vk::DeviceSize>(attachment_width * attachment_height * 4, result.memoryRequirements.size),
					.memoryTypeIndex = index
				});

				attachments[i] = device.createImage(image_create_info);

				device.bindImageMemory(attachments[i], memories[i], 0);

				attachment_views[i] = device.createImageView(vk::ImageViewCreateInfo{
					.image = attachments[i],
					.viewType = vk::ImageViewType::e2D,
					.format = attachment_formats[i],
					.subresourceRange = vk::ImageSubresourceRange{
						.aspectMask = vk::ImageAspectFlagBits::eColor,
						.baseMipLevel = 0u,
						.levelCount = 1u,
						.baseArrayLayer = 0u,
						.layerCount = 1u
					}
				});

			}
		}

	}

	///////////////////////////////////
	//       Graphics Pipeline       //
	///////////////////////////////////
	vk::ShaderModule vertex_module{};
	vk::ShaderModule fragment_module{};

	vk::Pipeline graphics_pipeline{};
	vk::PipelineLayout graphics_pipeline_layout{};

	{

		constexpr std::array dynamic_states = {
			vk::DynamicState::eScissor,
			vk::DynamicState::eViewport
		};

		const vk::StencilOpState stencil = {
			.failOp = vk::StencilOp::eKeep,
			.passOp = vk::StencilOp::eKeep,
			.depthFailOp = vk::StencilOp::eKeep,
			.compareOp = vk::CompareOp::eAlways,
			.compareMask = 0xffffffffu,
			.writeMask = 0xffffffffu,
			.reference = 0,
		};

		std::vector<vk::Format> color_blend_formats(std::size(attachments) - 1u);
		std::vector<vk::PipelineColorBlendAttachmentState> color_blend_attachments(std::size(attachments) - 1u);

		for (std::uint64_t i = 0; i < color_blend_formats.size(); i++) {

			color_blend_formats[i] = attachment_formats[i + 1];
			color_blend_attachments[i] = vk::PipelineColorBlendAttachmentState{
				.blendEnable = VK_TRUE,
				.srcColorBlendFactor = vk::BlendFactor::eSrcAlpha,
				.dstColorBlendFactor = vk::BlendFactor::eOneMinusSrcAlpha,
				.colorBlendOp = vk::BlendOp::eAdd,
				.srcAlphaBlendFactor = vk::BlendFactor::eSrcAlpha,
				.dstAlphaBlendFactor = vk::BlendFactor::eOneMinusSrcAlpha,
				.alphaBlendOp = vk::BlendOp::eAdd,
				.colorWriteMask = attachment_color_components[i]
			};

		}

		const vk::PipelineVertexInputStateCreateInfo vertex_input = {

		};

		const vk::PipelineDynamicStateCreateInfo dynamic_state = {
			.dynamicStateCount = static_cast<std::uint32_t>(dynamic_states.size()),
			.pDynamicStates = dynamic_states.data()
		};

		const vk::PipelineViewportStateCreateInfo viewport_state = {
			.viewportCount = 1u,
			.scissorCount = 1u
		};

		const vk::PipelineRasterizationStateCreateInfo rasterizer = {
			.cullMode = vk::CullModeFlagBits::eNone,
			.lineWidth = 1.0f
		};

		const vk::PipelineColorBlendStateCreateInfo color_blending = {
			.attachmentCount = static_cast<std::uint32_t>(color_blend_attachments.size()),
			.pAttachments = color_blend_attachments.data()
		};

		const vk::PipelineDepthStencilStateCreateInfo depth_stencil = {
			.depthTestEnable = VK_TRUE,
			.depthWriteEnable = VK_FALSE,
			.depthCompareOp = vk::CompareOp::eAlways,
			.depthBoundsTestEnable = VK_FALSE,
			.stencilTestEnable = VK_FALSE,
			.front = stencil,
			.back = stencil,
			.minDepthBounds = 0.0f,
			.maxDepthBounds = 1.0f
		};

		const vk::PipelineMultisampleStateCreateInfo multisampling = {
			.rasterizationSamples = attachment_samples[0],
			.sampleShadingEnable = VK_FALSE,
			.minSampleShading = 0.0f,
		};

		const vk::PipelineInputAssemblyStateCreateInfo input_assembly = {
			.topology = vk::PrimitiveTopology::eTriangleList
		};

		const auto vertex_spv = read_file(std::filesystem::current_path() / "shaders/mixed_sample_count/vert.spv");
		const auto fragment_spv = read_file(std::filesystem::current_path() / "shaders/mixed_sample_count/frag.spv");

		vertex_module = device.createShaderModule(vk::ShaderModuleCreateInfo{
			.codeSize = vertex_spv.size() * sizeof(std::uint32_t),
			.pCode = vertex_spv.data()
		});

		fragment_module = device.createShaderModule(vk::ShaderModuleCreateInfo{
			.codeSize = fragment_spv.size() * sizeof(std::uint32_t),
			.pCode = fragment_spv.data()
		});

		const std::vector<vk::SampleCountFlagBits> color_sample_counts(attachment_samples + 1, attachment_samples + std::size(attachment_samples));

		const vk::AttachmentSampleCountInfoNV sample_counts_info = {
			.colorAttachmentCount = static_cast<std::uint32_t>(color_sample_counts.size()),
			.pColorAttachmentSamples = color_sample_counts.data(),
			.depthStencilAttachmentSamples = attachment_samples[0]
		};

		const vk::PipelineRenderingCreateInfo rendering = {
			.pNext = &sample_counts_info,
			.viewMask = 0u,
			.colorAttachmentCount = static_cast<std::uint32_t>(color_blend_attachments.size()),
			.pColorAttachmentFormats = color_blend_formats.data(),
			.depthAttachmentFormat = attachment_formats[0],
			.stencilAttachmentFormat = vk::Format::eUndefined
		};

		const vk::PipelineShaderStageCreateInfo stages[] = {
			vk::PipelineShaderStageCreateInfo{
				.stage = vk::ShaderStageFlagBits::eVertex,
				.module = vertex_module,
				.pName = "main"
			},

			vk::PipelineShaderStageCreateInfo{
				.stage = vk::ShaderStageFlagBits::eFragment,
				.module = fragment_module,
				.pName = "main"
			}
		};

		graphics_pipeline_layout = device.createPipelineLayout(vk::PipelineLayoutCreateInfo{});

		const vk::GraphicsPipelineCreateInfo pipeline_info = {
			.pNext = &rendering,
			.stageCount = static_cast<std::uint32_t>(std::size(stages)),
			.pStages = stages,
			.pVertexInputState = &vertex_input,
			.pInputAssemblyState = &input_assembly,
			.pViewportState = &viewport_state,
			.pRasterizationState = &rasterizer,
			.pMultisampleState = &multisampling,
			.pDepthStencilState = &depth_stencil,
			.pColorBlendState = &color_blending,
			.pDynamicState = &dynamic_state,
			.layout = graphics_pipeline_layout
		};

		graphics_pipeline = device.createGraphicsPipeline({}, pipeline_info).value;

	}

	///////////////////////////////////
	//            Buffers            //
	///////////////////////////////////
	vk::DeviceMemory draw_buffer_memory{};
	vk::DeviceMemory count_buffer_memory{};

	vk::Buffer draw_buffer{};
	vk::Buffer count_buffer{};

	std::uint32_t* mapped_count_buffer{};
	vk::DrawIndirectCommand* mapped_draw_buffer{};

	{

		{

			const vk::BufferCreateInfo buffer_create_info = {
				.size = sizeof(vk::DrawIndirectCommand),
				.usage = vk::BufferUsageFlagBits::eIndirectBuffer
			};

			const vk::MemoryRequirements2 result = device.getBufferMemoryRequirements(vk::DeviceBufferMemoryRequirements{
				.pCreateInfo = &buffer_create_info
			});

			const auto index = find_memory_type(physical_device, result.memoryRequirements.memoryTypeBits, vk::MemoryPropertyFlagBits::eHostVisible |
																										   vk::MemoryPropertyFlagBits::eHostCoherent);

			draw_buffer_memory = device.allocateMemory(vk::MemoryAllocateInfo{
				.allocationSize = std::max<vk::DeviceSize>(attachment_width * attachment_height * 4, result.memoryRequirements.size),
				.memoryTypeIndex = index
			});

			draw_buffer = device.createBuffer(buffer_create_info);

			device.bindBufferMemory(draw_buffer, draw_buffer_memory, 0);
			mapped_draw_buffer = reinterpret_cast<vk::DrawIndirectCommand*>(device.mapMemory(draw_buffer_memory, 0, sizeof(vk::DrawIndirectCommand)));
		}

		{

			const vk::BufferCreateInfo buffer_create_info = {
				.size = sizeof(std::uint32_t),
				.usage = vk::BufferUsageFlagBits::eIndirectBuffer
			};

			const vk::MemoryRequirements2 result = device.getBufferMemoryRequirements(vk::DeviceBufferMemoryRequirements{
				.pCreateInfo = &buffer_create_info
			});

			const auto index = find_memory_type(physical_device, result.memoryRequirements.memoryTypeBits, vk::MemoryPropertyFlagBits::eHostVisible |
																										   vk::MemoryPropertyFlagBits::eHostCoherent);

			count_buffer_memory = device.allocateMemory(vk::MemoryAllocateInfo{
				.allocationSize = std::max<vk::DeviceSize>(attachment_width * attachment_height * 4, result.memoryRequirements.size),
				.memoryTypeIndex = index
			});

			count_buffer = device.createBuffer(buffer_create_info);

			device.bindBufferMemory(count_buffer, count_buffer_memory, 0);
			mapped_count_buffer = reinterpret_cast<std::uint32_t*>(device.mapMemory(count_buffer_memory, 0, sizeof(std::uint32_t)));
		}

	}

	(*mapped_count_buffer) = 1u;
	(*mapped_draw_buffer) = vk::DrawIndirectCommand{
		.vertexCount = 3u,
		.instanceCount = 1u,
		.firstVertex = 0u,
		.firstInstance = 0u
	};

	///////////////////////////////////
	//           Rendering           //
	///////////////////////////////////
	if (rdoc_api != nullptr) {
		rdoc_api->StartFrameCapture(RENDERDOC_DEVICEPOINTER_FROM_VKINSTANCE(static_cast<VkInstance>(instance)), nullptr);
	}

	command_buffer.begin(vk::CommandBufferBeginInfo{
		.flags = vk::CommandBufferUsageFlagBits::eOneTimeSubmit
	});

	for (std::uint64_t i = 1u; i < std::size(attachments); i++) {

		const vk::ImageMemoryBarrier barrier = {
			.srcAccessMask = vk::AccessFlagBits::eMemoryRead | vk::AccessFlagBits::eMemoryWrite,
			.dstAccessMask = vk::AccessFlagBits::eMemoryRead | vk::AccessFlagBits::eMemoryWrite | vk::AccessFlagBits::eColorAttachmentWrite,
			.oldLayout = vk::ImageLayout::eUndefined,
			.newLayout = vk::ImageLayout::eAttachmentOptimal,
			.srcQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED,
			.dstQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED,
			.image = attachments[i],
			.subresourceRange = vk::ImageSubresourceRange{
				.aspectMask = vk::ImageAspectFlagBits::eColor,
				.baseMipLevel = 0u,
				.levelCount = 1u,
				.baseArrayLayer = 0u,
				.layerCount = 1u
			}
		};

		command_buffer.pipelineBarrier(vk::PipelineStageFlagBits::eColorAttachmentOutput,
									   vk::PipelineStageFlagBits::eColorAttachmentOutput,
									   {}, {}, {}, barrier);

	}

	command_buffer.pipelineBarrier(vk::PipelineStageFlagBits::eEarlyFragmentTests |
								   vk::PipelineStageFlagBits::eLateFragmentTests,
								   vk::PipelineStageFlagBits::eEarlyFragmentTests |
								   vk::PipelineStageFlagBits::eLateFragmentTests,
								   {}, {}, {}, vk::ImageMemoryBarrier{
										.srcAccessMask = vk::AccessFlagBits::eMemoryRead | vk::AccessFlagBits::eMemoryWrite,
										.dstAccessMask = vk::AccessFlagBits::eMemoryRead | vk::AccessFlagBits::eMemoryWrite | 
															vk::AccessFlagBits::eDepthStencilAttachmentWrite | vk::AccessFlagBits::eDepthStencilAttachmentRead,
										.oldLayout = vk::ImageLayout::eUndefined,
										.newLayout = vk::ImageLayout::eAttachmentOptimal,
										.srcQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED,
										.dstQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED,
										.image = attachments[0],
										.subresourceRange = vk::ImageSubresourceRange{
											.aspectMask = vk::ImageAspectFlagBits::eDepth,
											.baseMipLevel = 0u,
											.levelCount = 1u,
											.baseArrayLayer = 0u,
											.layerCount = 1u
										}
									});

	const std::array color_attachments = {
		vk::RenderingAttachmentInfo{
			.imageView = attachment_views[1],
			.imageLayout = vk::ImageLayout::eAttachmentOptimal,
			.loadOp = vk::AttachmentLoadOp::eClear,
			.storeOp = vk::AttachmentStoreOp::eStore
		},

		vk::RenderingAttachmentInfo{
			.imageView = attachment_views[2],
			.imageLayout = vk::ImageLayout::eAttachmentOptimal,
			.loadOp = vk::AttachmentLoadOp::eClear,
			.storeOp = vk::AttachmentStoreOp::eStore
		},

		vk::RenderingAttachmentInfo{
			.imageView = attachment_views[3],
			.imageLayout = vk::ImageLayout::eAttachmentOptimal,
			.loadOp = vk::AttachmentLoadOp::eClear,
			.storeOp = vk::AttachmentStoreOp::eStore
		},

		vk::RenderingAttachmentInfo{
			.imageView = attachment_views[4],
			.imageLayout = vk::ImageLayout::eAttachmentOptimal,
			.loadOp = vk::AttachmentLoadOp::eClear,
			.storeOp = vk::AttachmentStoreOp::eStore
		}
	};

	const vk::RenderingAttachmentInfo depth_attachment = {
		.imageView = attachment_views[0],
		.imageLayout = vk::ImageLayout::eAttachmentOptimal,
		.loadOp = vk::AttachmentLoadOp::eClear,
		.storeOp = vk::AttachmentStoreOp::eStore
	};

	command_buffer.setViewport(0, vk::Viewport{
		.x = 0.0f,
		.y = 0.0f,
		.width = static_cast<float>(attachment_width),
		.height = static_cast<float>(attachment_height),
		.minDepth = 0.0f,
		.maxDepth = 1.0f
	});

	command_buffer.setScissor(0, vk::Rect2D{
		.offset = {},
		.extent = { attachment_width, attachment_height },
	});

	command_buffer.beginRendering(vk::RenderingInfo{
		.renderArea = vk::Rect2D{ 
			.offset = { 0, 0 },
			.extent = { attachment_width, attachment_height }
		},

		.layerCount = 1u,
		.viewMask = 0x0,
		.colorAttachmentCount = static_cast<std::uint32_t>(color_attachments.size()),
		.pColorAttachments = color_attachments.data(),
		.pDepthAttachment = &depth_attachment
	});

	command_buffer.bindPipeline(vk::PipelineBindPoint::eGraphics, graphics_pipeline);
	// command_buffer.draw(3, 1, 0, 0);
	command_buffer.drawIndirectCount(draw_buffer, 0, count_buffer, 0, 1, sizeof(vk::DrawIndirectCommand));
	command_buffer.endRendering();
	command_buffer.end();

	if (rdoc_api != nullptr) {
		rdoc_api->EndFrameCapture(RENDERDOC_DEVICEPOINTER_FROM_VKINSTANCE(static_cast<VkInstance>(instance)), nullptr);
	}

	constexpr vk::PipelineStageFlags wait = vk::PipelineStageFlagBits::eColorAttachmentOutput;

	graphics_queue.submit(vk::SubmitInfo{
		.pWaitDstStageMask = &wait,
		.commandBufferCount = 1u,
		.pCommandBuffers = &command_buffer
	});

	device.waitIdle();

	///////////////////////////////////
	//            Buffers            //
	///////////////////////////////////
	device.unmapMemory(draw_buffer_memory);
	device.unmapMemory(count_buffer_memory);
	device.destroy(draw_buffer);
	device.destroy(count_buffer);
	device.free(draw_buffer_memory);
	device.free(count_buffer_memory);

	///////////////////////////////////
	//       Graphics Pipeline       //
	///////////////////////////////////
	device.destroy(graphics_pipeline);
	device.destroy(graphics_pipeline_layout);
	device.destroy(fragment_module);
	device.destroy(vertex_module);

	///////////////////////////////////
	//         Attachments	         //
	///////////////////////////////////
	for (std::uint64_t i = 0; i < std::size(attachments); i++) {
		device.destroy(attachment_views[i]);
		device.destroy(attachments[i]);
		device.freeMemory(memories[i]);
	}

	///////////////////////////////////
	//         Command Buffer        //
	///////////////////////////////////
	device.free(command_pool, command_buffer);

	//////////////////////////////////
	//          Command Pool         //
	///////////////////////////////////
	device.destroy(command_pool);

	///////////////////////////////////
	//            Device             //
	///////////////////////////////////
	device.destroy();

	///////////////////////////////////
	//        Physcical Device       //
	///////////////////////////////////
	// nothing to delete

	///////////////////////////////////
	//           Instance            //
	///////////////////////////////////
#if defined(DEBUG) || defined(_DEBUG)
	instance.destroy(debug_messenger);
#endif
	instance.destroy();

	///////////////////////////////////
	//        Dynamic Dispatch       //
	///////////////////////////////////
	// nothing to delete

}