#include "VulkanHeaders.h"
#include "MixedSampleCountDynamicRendering.h"

#include "renderdoc_app.h"

#include <vector>

#define NOMINMAX
#include <Windows.h>
#undef MemoryBarrier

inline RENDERDOC_API_1_5_0* rdoc_api = nullptr;

void slow_pipeline_creation(const std::int32_t argc, const char** argv) {

	if (HMODULE mod = GetModuleHandleA("renderdoc.dll")) {
		pRENDERDOC_GetAPI RENDERDOC_GetAPI = (pRENDERDOC_GetAPI) GetProcAddress(mod, "RENDERDOC_GetAPI");
		RENDERDOC_GetAPI(eRENDERDOC_API_Version_1_5_0, (void**) &rdoc_api);
	}

	///////////////////////////////////
	//        Dynamic Dispatch       //
	///////////////////////////////////
	vk::DynamicLoader loader{};

	{

		PFN_vkGetInstanceProcAddr vkGetInstanceProcAddr = loader.getProcAddress<PFN_vkGetInstanceProcAddr>("vkGetInstanceProcAddr");

		VULKAN_HPP_DEFAULT_DISPATCHER.init(vkGetInstanceProcAddr);

	}

	///////////////////////////////////
	//           Instance            //
	///////////////////////////////////
	vk::Instance instance{};
	vk::DebugUtilsMessengerEXT debug_messenger{};

	{

		constexpr const char* extensions[] = {
			VK_EXT_DEBUG_UTILS_EXTENSION_NAME
		};

		constexpr const char* layers[] = {
			"VK_LAYER_KHRONOS_validation"
		};

		constexpr std::array enables = {
			vk::ValidationFeatureEnableEXT::eGpuAssisted,
			vk::ValidationFeatureEnableEXT::eSynchronizationValidation,
		};

		const vk::ApplicationInfo app_info = {
			.applicationVersion = VK_MAKE_API_VERSION(0, 0, 1, 0),
			.pEngineName = "NvidiaBugs",
			.engineVersion = VK_MAKE_API_VERSION(0, 0, 1, 0),
			.apiVersion = VK_API_VERSION_1_3
		};

		const vk::ValidationFeaturesEXT features = {
			.enabledValidationFeatureCount = static_cast<std::uint32_t>(enables.size()),
			.pEnabledValidationFeatures = enables.data(),
		};

		instance = vk::createInstance(vk::InstanceCreateInfo{
#if defined(DEBUG) || defined(_DEBUG)
			.pNext = &features,
#endif
			.pApplicationInfo = &app_info,
#if defined(DEBUG) || defined(_DEBUG)
			.enabledLayerCount = static_cast<std::uint32_t>(std::size(layers)),
			.ppEnabledLayerNames = layers,
			.enabledExtensionCount = static_cast<std::uint32_t>(std::size(extensions)),
			.ppEnabledExtensionNames = extensions,
#endif
		});

		VULKAN_HPP_DEFAULT_DISPATCHER.init(instance);
#if defined(DEBUG) || defined(_DEBUG)
		const vk::DebugUtilsMessengerCreateInfoEXT debug_info = {
			.messageSeverity = vk::DebugUtilsMessageSeverityFlagBitsEXT::eVerbose |
							   vk::DebugUtilsMessageSeverityFlagBitsEXT::eError |
							   vk::DebugUtilsMessageSeverityFlagBitsEXT::eWarning,
			.messageType = vk::DebugUtilsMessageTypeFlagBitsEXT::eGeneral |
						   vk::DebugUtilsMessageTypeFlagBitsEXT::ePerformance |
						   vk::DebugUtilsMessageTypeFlagBitsEXT::eValidation,
			.pfnUserCallback = debug_callback
		};

		debug_messenger = instance.createDebugUtilsMessengerEXT(debug_info);
#endif
	}

	///////////////////////////////////
	//        Physcical Device       //
	///////////////////////////////////
	vk::PhysicalDevice physical_device = instance.enumeratePhysicalDevices()[0];

	///////////////////////////////////
	//            Device             //
	///////////////////////////////////
	vk::Device device{};
	vk::Queue graphics_queue{};
	std::uint32_t graphics_queue_index{};

	{

		const std::vector<vk::QueueFamilyProperties> queue_families = physical_device.getQueueFamilyProperties();

		const auto findBestQueue = [&](const vk::QueueFlagBits bits) -> std::int32_t {
			const auto bestFinder = [&](const vk::QueueFamilyProperties& lhs,
										const vk::QueueFamilyProperties& rhs) noexcept -> bool {
				if (lhs.queueFlags & bits &&
					!(rhs.queueFlags & bits)) {
					return false;
				} else if (rhs.queueFlags & bits &&
						   !(lhs.queueFlags & bits)) {
					return true;
				}

				return std::popcount(static_cast<std::uint32_t>(lhs.queueFlags)) <
					   std::popcount(static_cast<std::uint32_t>(rhs.queueFlags));
			};

			const auto leastFinder = [&](const vk::QueueFamilyProperties& lhs,
										 const vk::QueueFamilyProperties& rhs) noexcept -> bool {
				if (lhs.queueFlags & bits &&
					!(rhs.queueFlags & bits)) {
					return false;
				} else if (rhs.queueFlags & bits &&
						   !(lhs.queueFlags & bits)) {
					return true;
				}

				return !(std::popcount(static_cast<std::uint32_t>(lhs.queueFlags)) <
						 std::popcount(static_cast<std::uint32_t>(rhs.queueFlags)));
			};

			if (bits == vk::QueueFlagBits::eGraphics) {
				const auto iterator = std::max_element(queue_families.begin(), queue_families.end(), bestFinder);
				return static_cast<std::int32_t>(std::distance(queue_families.begin(), iterator));
			} else {
				const auto iterator = std::max_element(queue_families.begin(), queue_families.end(), leastFinder);
				return static_cast<std::int32_t>(std::distance(queue_families.begin(), iterator));
			}
		};

		graphics_queue_index = findBestQueue(vk::QueueFlagBits::eGraphics);

		vk::StructureChain<vk::PhysicalDeviceFeatures2, vk::PhysicalDeviceVulkan12Features, vk::PhysicalDeviceVulkan13Features> features;

		std::get<0>(features).features.independentBlend = VK_TRUE;
		std::get<1>(features).drawIndirectCount = VK_TRUE;
		std::get<2>(features).dynamicRendering = VK_TRUE;
		std::get<2>(features).synchronization2 = VK_TRUE;

		constexpr float priority = 1.0f;

		const vk::DeviceQueueCreateInfo queue_create_info = {
			.queueFamilyIndex = graphics_queue_index,
			.queueCount = 1,
			.pQueuePriorities = &priority,
		};

		const vk::DeviceCreateInfo createInfo = {
			.pNext = &std::get<0>(features),
			.queueCreateInfoCount = 1,
			.pQueueCreateInfos = &queue_create_info,
			.enabledLayerCount = 0
		};

		device = physical_device.createDevice(createInfo);
		graphics_queue = device.getQueue(graphics_queue_index, 0);

		VULKAN_HPP_DEFAULT_DISPATCHER.init(instance, loader.getProcAddress<PFN_vkGetInstanceProcAddr>("vkGetInstanceProcAddr"),
										   device, loader.getProcAddress<PFN_vkGetDeviceProcAddr>("vkGetDeviceProcAddr"));

	}

	std::uint32_t attachment_width = 1280;
	std::uint32_t attachment_height = 720u;

	constexpr auto all_components = vk::ColorComponentFlagBits::eR | vk::ColorComponentFlagBits::eG | 
									vk::ColorComponentFlagBits::eB | vk::ColorComponentFlagBits::eA;
	constexpr vk::Format depth_format = vk::Format::eD32Sfloat;
	constexpr vk::Format color_format = vk::Format::eB10G11R11UfloatPack32;

	///////////////////////////////////
	//       Descriptor Layout       //
	///////////////////////////////////
	constexpr vk::DescriptorSetLayoutBinding descriptor_set_layout_binding[] = {
		vk::DescriptorSetLayoutBinding{
			.binding = 0u,
			.descriptorType = vk::DescriptorType::eStorageBuffer,
			.descriptorCount = 1,
			.stageFlags = vk::ShaderStageFlagBits::eVertex
		},

		vk::DescriptorSetLayoutBinding{
			.binding = 1u,
			.descriptorType = vk::DescriptorType::eStorageBuffer,
			.descriptorCount = 1u,
			.stageFlags = vk::ShaderStageFlagBits::eFragment
		},

		vk::DescriptorSetLayoutBinding{
			.binding = 2u,
			.descriptorType = vk::DescriptorType::eCombinedImageSampler,
			.descriptorCount = 256u * 9u,
			.stageFlags = vk::ShaderStageFlagBits::eFragment
		}
	};

	vk::DescriptorSetLayout descriptor_set_layout = device.createDescriptorSetLayout(vk::DescriptorSetLayoutCreateInfo{
		.bindingCount = 3u,
		.pBindings = descriptor_set_layout_binding
	});

	///////////////////////////////////
	//       Graphics Pipeline       //
	///////////////////////////////////
	vk::ShaderModule vertex_module{};
	vk::ShaderModule fragment_module{};

	vk::Pipeline graphics_pipeline{};
	vk::PipelineLayout graphics_pipeline_layout{};

	{

		constexpr std::array dynamic_states = {
			vk::DynamicState::eScissor,
			vk::DynamicState::eViewport
		};

		const vk::StencilOpState stencil = {
			.failOp = vk::StencilOp::eKeep,
			.passOp = vk::StencilOp::eKeep,
			.depthFailOp = vk::StencilOp::eKeep,
			.compareOp = vk::CompareOp::eAlways,
			.compareMask = 0xffffffffu,
			.writeMask = 0xffffffffu,
			.reference = 0,
		};

		constexpr vk::PipelineColorBlendAttachmentState color_blend_attachment = vk::PipelineColorBlendAttachmentState{
			.blendEnable = VK_FALSE,
			.srcColorBlendFactor = vk::BlendFactor::eSrcAlpha,
			.dstColorBlendFactor = vk::BlendFactor::eOneMinusSrcAlpha,
			.colorBlendOp = vk::BlendOp::eAdd,
			.srcAlphaBlendFactor = vk::BlendFactor::eSrcAlpha,
			.dstAlphaBlendFactor = vk::BlendFactor::eOneMinusSrcAlpha,
			.alphaBlendOp = vk::BlendOp::eAdd,
			.colorWriteMask = all_components
		};

		const vk::PipelineVertexInputStateCreateInfo vertex_input = {

		};

		const vk::PipelineDynamicStateCreateInfo dynamic_state = {
			.dynamicStateCount = static_cast<std::uint32_t>(dynamic_states.size()),
			.pDynamicStates = dynamic_states.data()
		};

		const vk::PipelineViewportStateCreateInfo viewport_state = {
			.viewportCount = 1u,
			.scissorCount = 1u
		};

		const vk::PipelineRasterizationStateCreateInfo rasterizer = {
			.cullMode = vk::CullModeFlagBits::eNone,
			.lineWidth = 1.0f
		};

		const vk::PipelineColorBlendStateCreateInfo color_blending = {
			.attachmentCount = 1u,
			.pAttachments = &color_blend_attachment
		};

		const vk::PipelineDepthStencilStateCreateInfo depth_stencil = {
			.depthTestEnable = VK_TRUE,
			.depthWriteEnable = VK_FALSE,
			.depthCompareOp = vk::CompareOp::eAlways,
			.depthBoundsTestEnable = VK_FALSE,
			.stencilTestEnable = VK_FALSE,
			.front = stencil,
			.back = stencil,
			.minDepthBounds = 0.0f,
			.maxDepthBounds = 1.0f
		};

		const vk::PipelineMultisampleStateCreateInfo multisampling = {
			.rasterizationSamples = vk::SampleCountFlagBits::e1,
			.sampleShadingEnable = VK_FALSE,
			.minSampleShading = 0.0f,
		};

		const vk::PipelineInputAssemblyStateCreateInfo input_assembly = {
			.topology = vk::PrimitiveTopology::eTriangleList
		};

		const auto vertex_spv = read_file(std::filesystem::current_path() / "shaders/slow_pipeline_creation/vert.spv");
		const auto fragment_spv = read_file(std::filesystem::current_path() / "shaders/slow_pipeline_creation/frag.spv");

		vertex_module = device.createShaderModule(vk::ShaderModuleCreateInfo{
			.codeSize = vertex_spv.size() * sizeof(std::uint32_t),
			.pCode = vertex_spv.data()
		});

		fragment_module = device.createShaderModule(vk::ShaderModuleCreateInfo{
			.codeSize = fragment_spv.size() * sizeof(std::uint32_t),
			.pCode = fragment_spv.data()
		});

		const vk::PipelineRenderingCreateInfo rendering = {
			.viewMask = 0u,
			.colorAttachmentCount = 1u,
			.pColorAttachmentFormats = &color_format,
			.depthAttachmentFormat = depth_format,
			.stencilAttachmentFormat = vk::Format::eUndefined
		};

		const vk::PipelineShaderStageCreateInfo stages[] = {
			vk::PipelineShaderStageCreateInfo{
				.stage = vk::ShaderStageFlagBits::eVertex,
				.module = vertex_module,
				.pName = "main"
			},

			vk::PipelineShaderStageCreateInfo{
				.stage = vk::ShaderStageFlagBits::eFragment,
				.module = fragment_module,
				.pName = "main"
			}
		};

		graphics_pipeline_layout = device.createPipelineLayout(vk::PipelineLayoutCreateInfo{
			.setLayoutCount = 1u,
			.pSetLayouts = &descriptor_set_layout
		});

		const vk::GraphicsPipelineCreateInfo pipeline_info = {
			.pNext = &rendering,
			.stageCount = static_cast<std::uint32_t>(std::size(stages)),
			.pStages = stages,
			.pVertexInputState = &vertex_input,
			.pInputAssemblyState = &input_assembly,
			.pViewportState = &viewport_state,
			.pRasterizationState = &rasterizer,
			.pMultisampleState = &multisampling,
			.pDepthStencilState = &depth_stencil,
			.pColorBlendState = &color_blending,
			.pDynamicState = &dynamic_state,
			.layout = graphics_pipeline_layout
		};

		graphics_pipeline = device.createGraphicsPipeline({}, pipeline_info).value;

	}

	///////////////////////////////////
	//       Graphics Pipeline       //
	///////////////////////////////////
	device.destroy(graphics_pipeline);
	device.destroy(graphics_pipeline_layout);
	device.destroy(fragment_module);
	device.destroy(vertex_module);

	///////////////////////////////////
	//       Descriptor Layout       //
	///////////////////////////////////
	device.destroy(descriptor_set_layout);

	///////////////////////////////////
	//            Device             //
	///////////////////////////////////
	device.destroy();

	///////////////////////////////////
	//        Physcical Device       //
	///////////////////////////////////
	// nothing to delete

	///////////////////////////////////
	//           Instance            //
	///////////////////////////////////
#if defined(DEBUG) || defined(_DEBUG)
	instance.destroy(debug_messenger);
#endif
	instance.destroy();

	///////////////////////////////////
	//        Dynamic Dispatch       //
	///////////////////////////////////
	// nothing to delete

}